import React, { Fragment } from "react";
import { connect } from "react-redux";
import { compose } from "recompose";
import { withFirebase } from "../Firebase";
import { withRouter } from "react-router-dom";

import MapBase from "./MapBase";
import LotPopup from "./LotPopup";
// import StatusFilters from "../StatusFilters/StatusFilters";
import mapboxgl from "mapbox-gl";

import { FILTER } from "../../constants/lot";
import * as ROUTES from "../../constants/routes";

import { updateAuthUserBalance } from "../../actions/user";
import { requestListRefresh } from "../../actions/lot";

import * as BIT from "../../utils/Bitwise";
import { async } from "@firebase/util";

import * as UTILS from "./utils";

import "./Map.scss";

// const FILL_OPACITY = 0.2;

// const LINE_WIDTH = 1;

// const LOT_PADDING_K = 0.03;
// const LOT_PADDING_K2 = 0.67;

const GP_ACCURACY_THRESHOLD = 1000;

const CELL_SIZE = 6.77;

// const EMPTY_COLLECTION = {
//   type: "FeatureCollection",
//   features: []
// };

class Map extends MapBase {
  constructor(props) {
    super(props);

    const lastPosition = JSON.parse(
      localStorage.getItem("avvy:map:lastPosition") ||
        `{"lng":37.635308,"lat":55.737694}`
    );

    this.state = {
      ...this.state,
      ...lastPosition,
      zoom: 16,
      selectedLots: false,
      clusterPos: [],
      cells: {}
    };

    this.markers = {};
    for (let key in UTILS.bubbles_colors) {
      this.markers[key] = [];
    }

    this.popup = null;

    this.updateTimeout = null;
  }

  componentDidUpdate(prevProps, prevState) {
    const { filter } = this.props;
    if (filter !== prevProps.filter) this.updateByFilterState();
  }

  calcScale() {
    const { zoom } = this.state;
    return zoom >= 16 ? 1 : Math.pow(2, Math.ceil(17 - zoom));
  }

  drawCells() {
    this.closePopup();
    const { cells } = this.state;
    const { authUser } = this.props;
    const { zoom } = this.state;
    // const paddingLeft = zoom < 16 ? LOT_PADDING_K2 : LOT_PADDING_K;
    // const paddingRight = zoom < 16 ? LOT_PADDING_K2 : LOT_PADDING_K;

    for (let fc in this.markers) {
      for (var i = this.markers[fc].length - 1; i >= 0; i--) {
        this.markers[fc][i].remove();
      }
      this.markers[fc] = [];
    }

    for (let key in cells) {
      const cell = cells[key];

      const count = Object.keys(cell.cells).length;
      const inner = count;
      const feature = {
        type: "Feature",
        properties: {
          iconSize: [50, 50],
          count,
          id: key
        },
        geometry: {
          type: "Point",
          position: [cell.position._longitude, cell.position._latitude]
        }
      };
      const type = UTILS.getLotType(cell, authUser);
      const position = feature.geometry.position;
      const [width, height] = feature.properties.iconSize;
      const listeners = {
        dblclick: this.onMapDblclick,
        mouseup: this.onMapClick
      };

      this.markers[type].push(
        this.bubble({
          type,
          position,
          feature,
          inner,
          width,
          height,
          listeners
        })
      );
    }

    this.updateByFilterState();
  }

  updateByFilterState() {
    const { filter } = this.props;
    for (let key in UTILS.bubbles_colors) {
      if (
        key === "own" ||
        key === "rest" ||
        (key === "ownForSale" &&
          BIT.isset(filter, FILTER.FOR_SALE + FILTER.OWN_SHIFT)) ||
        (key === "ownForRent" &&
          BIT.isset(filter, FILTER.FOR_RENT + FILTER.OWN_SHIFT)) ||
        (key === "ownRented" &&
          BIT.isset(filter, FILTER.RENTED + FILTER.OWN_SHIFT)) ||
        (key === "ownLeased" && BIT.isset(filter, FILTER.LEASED)) ||
        (key === "ownAuction" &&
          BIT.isset(filter, FILTER.AUCTION + FILTER.OWN_SHIFT)) ||
        ((key === "forSale" || key === "forSaleOwned") &&
          BIT.isset(filter, FILTER.FOR_SALE)) ||
        // (key === "forSale" && BIT.isset(filter, FILTER.FREE_SALE)) ||
        // (key === "forSaleOwned" && BIT.isset(filter, FILTER.FOR_SALE)) ||
        (key === "forRent" && BIT.isset(filter, FILTER.FOR_RENT)) ||
        (key === "auction" && BIT.isset(filter, FILTER.AUCTION))
      ) {
        for (let i = this.markers[key].length - 1; i >= 0; i--) {
          this.markers[key][i].addTo(this.map);
        }
      } else {
        for (let i = this.markers[key].length - 1; i >= 0; i--) {
          this.markers[key][i].remove();
        }
      }
    }
  }

  updateCellsInView(callback) {
    if (this.updateTimeout) {
      clearTimeout(this.updateTimeout);
    }
    this.updateTimeout = setTimeout(() => {
      this._updateCellsInView(callback);
    }, 100);
  }

  _updateCellsInView(callback) {
    const { firebase } = this.props;
    const { lng, lat } = this.map.getCenter();
    const zoom = this.map.getZoom().toFixed(2);

    this.setState({
      lng: lng,
      lat: lat,
      zoom: zoom,
      loading: true
    });

    let scale = this.calcScale();

    if (scale > 1) {
      firebase
        .cellsByCoord({
          longitude: lng,
          latitude: lat,
          width: 100 * scale, //120
          height: 100 * scale, //70
          scale: scale
        })
        .then(data => {
          const cells = JSON.parse(data.data.lots);
          this.setState({ cells });
          this.drawCells();
        })
        .finally(() => {
          this.setState({ loading: false });
        });
    } else {
      firebase
        .lotsByCoords({
          longitude: lng,
          latitude: lat,
          width: 100 * scale, //120
          height: 100 * scale, //70
          scale: scale
        })
        .then(data => {
          const cells = JSON.parse(data.data.lots);
          this.setState({ cells });
          this.drawCells();
        })
        .finally(() => {
          this.setState({ loading: false });
        });
    }
    this.controlHeight();
  }

  onMapLoad() {
    super.onMapLoad();
    var context = this;

    this.setState({ selectedLots: false });

    var content = document.getElementById("lot-popup-content");
    this.popup = new mapboxgl.Popup({ closeOnClick: false, closeButton: true })
      .setDOMContent(content)
      .setLngLat([0, 0])
      .addTo(this.map);

    this.closePopup();
    this.popup.remove = function() {
      context.popup._container.classList.add("hide");
    };
    // this.popup.remove();

    this.map.on("zoom", this.onMapZoom.bind(this));

    if (navigator.geolocation) {
      navigator.geolocation.getCurrentPosition(
        position => {
          const { longitude, latitude, accuracy } = position.coords;
          const pos = { lng: longitude, lat: latitude };
          if (accuracy <= GP_ACCURACY_THRESHOLD) {
            const rfn = function(cnt = 0) {
              if (this.state.loading) {
                setTimeout(() => {
                  rfn.call(this, cnt++);
                }, 100);
              } else {
                localStorage.setItem(
                  "avvy:map:lastPosition",
                  JSON.stringify(pos)
                );
                this.map.setCenter(pos);
              }
            };
            rfn.call(this);
          }
        },
        positionError => {
          console.warn("POSITION ERROR", positionError);
        }
      );
    }

    this.updateCellsInView();
  }

  onMapZoom() {
    this.closePopup();
  }

  onMapClick(e) {
    if (this.state.loading) return;

    const { firebase } = this.props;

    const pos = e.currentTarget.feature.geometry.position;
    const uid = e.currentTarget.feature.properties.id;

    let scale = this.calcScale();

    this.setState({
      selectedLots: false,
      clusterPos: pos
    });

    this.showPopup();

    if (scale > 1) {
      firebase
        .lotsByCoords({
          longitude: pos[1],
          latitude: pos[0],
          width: CELL_SIZE * scale, //120
          height: CELL_SIZE * scale //70
        })
        .then(cells => {
          this.setState({ selectedLots: JSON.parse(cells.data.lots) });
        });
    } else {
      firebase.lotByUid(uid).then(cells => {
        this.setState({ selectedLots: JSON.parse(cells.data.lots) });
      });
    }
  }

  onMapDblclick(e) {
    if (this.state.loading) return;

    e.preventDefault();
    this.closePopup();
    let scale = this.calcScale();
    if (scale === 1) {
      const { firebase } = this.props;

      const pos = e.currentTarget.feature.geometry.position;
      const uid = e.currentTarget.feature.properties.id;
      firebase.lotByUid(uid).then(cells => {
        const id = Object.keys(JSON.parse(cells.data.lots))[0];
        this.props.history.push(ROUTES.LOT.replace(/:[^:]+$/, id));
      });
    }
  }

  onMapMoveEnd() {
    const { lng, lat, zoom } = this.state;

    this.setState({
      lng: lng,
      lat: lat,
      zoom: zoom
    });
    localStorage.setItem("avvy:map:lastPosition", JSON.stringify({ lng, lat }));

    this.updateCellsInView();
    super.onMapMoveEnd();
  }

  showPopup() {
    const { clusterPos } = this.state;

    this.popup.setLngLat(clusterPos);

    this.popup._container.classList.remove("hide");
  }

  closePopup() {
    this.popup._container.classList.add("hide");
  }

  onBuy() {
    const {
      firebase,
      updateAuthUserBalance,
      authUser,
      requestListRefresh
    } = this.props;
    this.closePopup();
    this.updateCellsInView();
    firebase
      .user(authUser.uid)
      .once("value")
      .then(snapshot => {
        const balance = snapshot.val().Balance;
        updateAuthUserBalance(balance);
        requestListRefresh(3);
      });
  }

  render() {
    const { zoom, selectedLots } = this.state;

    return (
      <Fragment>
        <LotPopup
          lots={selectedLots}
          canBuy={zoom >= 16}
          callback={this.onBuy.bind(this)}
        />
        {/* <StatusFilters /> */}
        {super.render()}
      </Fragment>
    );
  }
}

const mapStateToProps = state => {
  const { authUser } = state.session;
  const { filter } = state.lot;
  return { authUser, filter };
};
const mapDispatchToProps = dispatch => ({
  updateAuthUserBalance: value => dispatch(updateAuthUserBalance(value)),
  requestListRefresh: value => dispatch(requestListRefresh(value))
});

export default compose(
  withRouter,
  withFirebase,
  connect(
    mapStateToProps,
    mapDispatchToProps
  )
)(Map);
